
open Terms;;
open Subst;;
open Mterms;;
open Monitor;;

(********** Lazy **********)

(* Lookup in the context *)
let clos_rel e i =
inc "expand";
  match expand_rel e i with
      Inl(0,mt) -> inc "defined0"; mt
    | Inl(n,mt) -> inc "defined";
        let (no,ft) = lift_fterm n mt in { norm=no; term=ft }
    | Inr n -> inc "record"; {norm=Norm; term= Frel n}
;;

type stk =
    Ztop
  | Zapp of mterm * stk
  | Zlam of mterm subs * lambda * stk
  | Zlift of int * stk
  | Zupdate of mterm * stk
  | Zarg of mterm * stk
;;

let zup m s =
  if m.norm = Red then Zupdate(m,s) else s
;;

let zlift n s =
  match (n,s) with
      (0,_) -> s
    | (_,Zlift(k,s)) -> Zlift(n+k,s)
    | _ -> Zlift(n,s)
;;

let ge = ref subs_id;;
let gs = ref Ztop;;


let rec knh m =
  if is_val m then m else
  match m.term with
  | Fapp(a,b) ->
inc "fapp";
      gs := Zarg (b, zup m !gs);
      knh a
  | Flift(k,a) ->
      gs := zlift k !gs;
      knh a
  | Frel i ->
      set_norm m;
      m
  | Flam (bd,e,f) ->
      set_whnf m;
      m
  | Fclos(e,f) ->
inc "unfreeze";
      ge := e;
      gs := Zupdate(m,!gs);
      knht f

and knht = function
    App(a,b) ->
inc "app";
      gs := Zarg (mk_clos !ge b, !gs);
      knht a
  | Lam f ->
inc "record";
      {norm=Whnf; term=Flam(mk_clos (subs_lift !ge) f, !ge, f)}
  | Rel i -> knh (clos_rel !ge i)
;;


(* *)

let get_arg h =
  let rec get_arg_rec depth = function
      Zarg(arg,s) ->
        gs := s;
        Some(depth,arg)
    | Zlift(k,s) -> get_arg_rec (k+depth) s
    | Zupdate(m,s) ->
        update m (lift_fterm depth h);
        get_arg_rec depth s
    | s ->
        gs := zlift depth s;
        None
  in get_arg_rec 0 !gs
;;

let rec knr m =
  match m.term with
  | Flam (bd,e,f) ->
      (match get_arg m with
           Some (depth,arg) ->
inc "beta";
             knit (subs_shift_cons(depth,e,arg)) f
         | None ->
             if is_val m then m else
               (gs := Zlam (e,f,!gs);
                kni bd))
  | _ -> m

and kni m = knr (knh m)
and knit e t =
  ge := e;
  knr (knht t)
;;


let rec zip m =
  match !gs with
    Ztop -> m
  | Zapp(a,s) ->
      gs := s;
      zip {norm=Norm; term=Fapp(a,m)}
  | Zlam(e,f,s) ->
      gs := s;
      zip {norm=Norm; term=Flam(m,e,f)}
  | Zlift(k,s) ->
      gs := s;
      zip {norm=Norm; term=Flift(k,m)}
  | Zupdate(m',s) ->
      gs := s;
      zip (update m' (Norm, m.term));
  | Zarg(a,s) ->
      gs := Zapp(m,s);
      if is_val a then zip a else a
;;

let rec walk m =
  let rm = zip (kni m) in
    match !gs with
        Ztop -> rm
      | _ -> walk rm
;;

let to_lambda t =
  let rec lbda_rec lfts m =
    match m.term with
        Frel i -> Rel (reloc_rel lfts i)
      | Fapp(a,b) -> App(lbda_rec lfts a, lbda_rec lfts b)
      | Flam(u,_,_) -> Lam(lbda_rec (llift lfts) u)
      | Flift (k,m) -> lbda_rec (lshift k lfts) m
      | Fclos _ -> failwith "lbda_rec"
  in lbda_rec lid t
;;


let kl t =
init();
  gs := Ztop;
  walk (knit subs_id t)
;;
let skl t = kl t; stat();;

let nlazy t = to_lambda (kl t)
;;
