
open Terms;;
open Format;;

let church t =
  let rec is_church n = function
      App(Rel 1,t) -> is_church (n+1) t
    | Rel 0 -> Some n
    | _ -> None
  in is_church 0 t
;;

let pr_church n =
  print_string "'";
  print_int n
;;

let pr_lam_box n f x =
  if n=0 then f x
  else begin
    open_hovbox 1;
    print_string "[";
    print_int n;
    print_string "]";
    print_break 0 0;
    f x;
    close_box()
  end
;;

(* Pretty-printer for pure lambda terms *)
(* Also recognizes Church numbers and tuples *)

let rec prl = function
    Rel n ->
      print_string "#";
      print_int n
  | Lam t -> pr_lam 1 t
  | App(a,b) -> pr_app [b] a

and pr_lam n = function
  | Lam t -> pr_lam (n+1) t
  | App(App(Rel 0,a),b) -> pr_lam_box (n-1) (pr_pair [b]) a
  | f ->
      if n>=2 then
        (match church f with
             Some k -> pr_lam_box (n-2) pr_church k
           | None -> pr_lam_box n prl f)
      else pr_lam_box n prl f

and pr_app stk = function
    App(a,b) -> pr_app (b::stk) a
  | h ->
      open_hovbox 2;
      print_string "(";
      prl h;
      List.iter
        (fun x ->
           print_break 1 0;
           prl x)
        stk;
      print_string ")";
      close_box()

and pr_pair stk = function
    Lam(App(App(Rel 0,a),b)) -> pr_pair (b::stk) a
  | h ->
      open_hvbox 1;
      print_string "<";
      prl h;
      List.iter
        (fun x ->
           print_string ",";
           print_break 1 0;
           prl x)
        stk;
      print_string ">";
      close_box()

;;

