
open Terms;;


let rec strong f t =
  match (f t) with
    Rel i -> Rel i
  | Lam u -> Lam (strong f u)
  | App(a,b) ->
      App(strong f a, strong f b)
;;


(* A few naive implementations *)

let rec subst_rec n k = function
    Rel i ->
      if i<k then Rel i
      else if i=k then (lift k 0 n)
      else Rel (i-1)
  | Lam t -> Lam (subst_rec n (k+1) t)
  | App(a,b) -> App(subst_rec n k a, subst_rec n k b)
;;
let subst n t = subst_rec n 0 t
;;

(* An awfull one *)
let rec norm0 = function
  Rel i    -> Rel i
| Lam t    -> Lam (norm0 t)
| App(a,b) ->
    (match (norm0 a) with
       Lam f -> norm0 (subst b f)
     | na -> App(na, norm0 b))
;;

(* Call-by-name *)
let rec nrm_stk = function
  (App(a,b), s)   -> nrm_stk (a, (b::s))
| (Lam f, (b::s)) ->
    nrm_stk (subst b f, s)
| (Lam t, [])     -> Lam (norm' t)
| (Rel i, s)      ->
    app_list (Rel i) (List.map norm' s)
and norm' t = nrm_stk (t,[])
;;

let rec nrm_stk = function
  (App(a,b), s)   -> nrm_stk (a, (b::s))
| (Lam f, (b::s)) ->
    nrm_stk (subst b f, s)
| (Lam t, [])     -> Lam (norm'' t)
| (Rel i, s)      -> nrm_app (Rel i, s)

and nrm_app = function
  (t,[]) -> t
| (t,(a::l)) -> nrm_app (App(t,norm'' a), l)

and norm'' t = nrm_stk (t,[])
;;

(* Call-by-value *)
let rec cbv_norm = function
  Rel i    -> Rel i
| Lam t    -> Lam (cbv_norm t)
| App(a,b) ->
    (match (cbv_norm a) with
       Lam f -> cbv_norm (subst (cbv_norm b) f)
     | na -> App(na, cbv_norm b))
;;

let rec cbv_whd = function
  Rel i    -> Rel i
| Lam t    -> Lam t
| App(a,b) ->
    (match (cbv_whd a) with
       Lam f -> cbv_whd (subst (cbv_whd b) f)
     | na -> App(na, b))
;;
let cbv_norm' = strong cbv_whd;;



(* Cf Coq V5.10 *)
let rec hnf = function
  App(a,b) ->
    (match (hnf a) with
       Lam f -> hnf (subst b f)
     | a' -> App(a',b))
| t -> t
;;

let norm = strong hnf;;

